"""This module provides functionality to anonymize sensitive information in text."""

import logging
import re
import subprocess

import spacy


class Anon:
    """Class to handle anonymization of sensitive information in text."""

    def __init__(self) -> None:
        """Initialize the Anon class and load the spaCy model."""
        try:
            self.nlp = spacy.load("en_core_web_sm")
        except Exception as e:
            logging.error(f"Error loading spacy model: {e}")
            subprocess.call("python -m spacy download en_core_web_sm", shell=True)
            self.nlp = spacy.load("en_core_web_sm")

        # Precompile regular expressions for better performance
        self.email_regex = re.compile(
            r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
        )
        self.credit_card_regex = re.compile(r"\b(?:\d[ -]*?){13,16}\b")

        # Default mapping of Spacy NER tags to custom sensitive information types
        self.info_type_mapping = {
            "PERSON": "PERSON",
            "CARDINAL": "NUMBER",  # Compromise for phone numbers
            "QUANTITY": "NUMBER",  # Compromise for phone numbers
            "GPE": "LOCATION",
        }



from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from enum import unique
else:
    unique, = jac_import('enum', 'py', items={'unique': None})
if typing.TYPE_CHECKING:
    from typing import Any
else:
    Any, = jac_import('typing', 'py', items={'Any': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})

@unique
class MessageType(Enum):
    TEXT = 'TEXT'
    MEDIA = 'MEDIA'
    MULTI = 'MULTI'
    SILENCE = 'SILENCE'

class InteractionMessage(Obj):
    message_type: str = field(None)
    content: Any = field(None)
    meta: dict = field(gen=lambda: {})

    def load(self, data: dict) -> None:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    setattr(self, attr, data[attr])

    def get_type(self) -> str:
        return self.message_type

    def set_meta(self, key: str, value: any) -> None:
        self.meta[key] = value

    def get_meta(self, key: str) -> None:
        return self.meta.get(key, None)

    def get_content(self) -> None:
        return self.content

    def has_content(self) -> bool:
        if self.content:
            return True
        return False

    def export(self, ignore_keys: list=JacList(['__jac__'])) -> None:
        node_export = Utils.export_to_dict(self, ignore_keys)
        return node_export

class SilentInteractionMessage(InteractionMessage, Obj):
    message_type: str = field(gen=lambda: MessageType.SILENCE.value)
    content: str = field('...')

class TextInteractionMessage(InteractionMessage, Obj):
    message_type: str = field(gen=lambda: MessageType.TEXT.value)
    content: str = field('')

class MediaInteractionMessage(InteractionMessage, Obj):
    message_type: str = field(gen=lambda: MessageType.MEDIA.value)
    mime: str = field('')
    data: Any = field(None)
    content: str = field('')

    def has_content(self) -> bool:
        if self.data:
            return True
        return False

class MultiInteractionMessage(InteractionMessage, Obj):
    message_type: str = field(gen=lambda: MessageType.MULTI.value)
    content: list[dict] = field(gen=lambda: JacList([]))

    def load(self, data: dict) -> None:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    setattr(self, attr, data[attr])

    def add_interaction_message(self, message: InteractionMessage) -> None:
        if not isinstance(message, MultiInteractionMessage):
            self.content.append(message.export())

    def clear_interaction_messages(self) -> None:
        self.content = JacList([])

    def has_content(self) -> bool:
        if type(self.content) == list:
            if len(self.content) > 0:
                return True
        return False

    def get_content(self) -> None:
        content = ''
        for content_item in self.content:
            content = f"{content} \\n {content_item.get('content')}"
        return content

    def get_content_items(self) -> None:
        content_items = JacList([])
        if self.content:
            for content_item in self.content:
                message = SilentInteractionMessage()
                if content_item.get('message_type') == MessageType.MULTI.value:
                    message = MultiInteractionMessage()
                elif content_item.get('message_type') == MessageType.MEDIA.value:
                    message = MediaInteractionMessage()
                elif content_item.get('message_type') == MessageType.TEXT.value:
                    message = TextInteractionMessage()
                message.load(content_item)
                content_items.append(message)
        return content_items

class InteractionResponse(Obj):
    session_id: str = field('')
    message_type: str = field(gen=lambda: MessageType.TEXT.value)
    message: dict = field(gen=lambda: {})
    tokens: int = field(0)

    def load(self, data: dict) -> None:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    setattr(self, attr, data[attr])

    def set_message(self, message: InteractionMessage) -> None:
        if message:
            self.message = message.export()
            self.message_type = message.get_type()

    def set_tokens(self, tokens: int) -> None:
        self.tokens = tokens

    def get_type(self) -> None:
        return self.message_type

    def get_message(self) -> InteractionMessage:
        message = None
        if self.message:
            if self.message_type == MessageType.MULTI.value:
                message = MultiInteractionMessage()
            elif self.message_type == MessageType.MEDIA.value:
                message = MediaInteractionMessage()
            elif self.message_type == MessageType.TEXT.value:
                message = TextInteractionMessage()
            else:
                message = SilentInteractionMessage()
            message.load(self.message)
        return message

    def export(self, ignore_keys: list=JacList(['__jac__'])) -> None:
        node_export = Utils.export_to_dict(self, ignore_keys)
        return node_export


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from graph_walker import graph_walker
else:
    graph_walker, = jac_import('graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})

class export_agent(graph_walker, Walker):

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        if self.reporting:
            Jac.report(here.export_descriptor())


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from app import App
else:
    App, = jac_import('app', items={'App': None})
if typing.TYPE_CHECKING:
    from agents import Agents
else:
    Agents, = jac_import('agents', items={'Agents': None})

class graph_walker(Walker):
    agent_id: str = field('')
    reporting: bool = field(True)
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = True

    def export(self, ignore_keys: list=JacList(['__jac__'])) -> None:
        node_export = Utils.export_to_dict(self, ignore_keys)
        return node_export

    def update(self, data: dict={}) -> graph_walker:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    self.attr = data[attr]
        self.postupdate()
        return self

    def postupdate(self) -> None:
        pass

    @with_entry
    def on_root(self, here: Root) -> None:
        if not self.visit(here.refs().filter(App, None)):
            self.logger.debug('app node created')
            app_node = here.connect(App())
            self.visit(app_node)

    @with_entry
    def on_app(self, here: App) -> None:
        if not self.visit(here.refs().filter(Agents, None)):
            self.logger.debug('agents node created')
            agents_node = here.connect(Agents())
            self.visit(agents_node)

    @with_entry
    def on_agents(self, here: Agents) -> None:
        if self.agent_id:
            try:
                if (agent_node := jobj(id=self.agent_id)):
                    if agent_node.published:
                        self.visit(agent_node)
            except Exception as e:
                Jac.get_context().status = 400
                Jac.report('Invalid agent id')
                return self.disengage()


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_walker import graph_walker
else:
    graph_walker, = jac_import('jivas.agent.core.graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.frame import Frame
else:
    Frame, = jac_import('jivas.agent.memory.frame', items={'Frame': None})

class get_interactions(graph_walker, Walker):
    interactions: dict = field(gen=lambda: {})
    session_id: str = field('')
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Memory, None))

    @with_entry
    def on_memory(self, here: Memory) -> None:
        if self.session_id:
            self.visit(here.refs().filter(Frame, None).filter(None, lambda item: item.session_id == self.session_id))
        else:
            self.visit(here.refs().filter(Frame, None))

    @with_entry
    def on_frame(self, here: Frame) -> None:
        interaction_list = here.get_interactions()
        self.interactions[here.session_id] = interaction_list

    @with_exit
    def on_exit(self, here) -> None:
        Jac.report(self.interactions)


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from datetime import datetime, timedelta
else:
    datetime, timedelta = jac_import('datetime', 'py', items={'datetime': None, 'timedelta': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_walker import graph_walker
else:
    graph_walker, = jac_import('jivas.agent.core.graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jac_cloud.core.architype import NodeAnchor
else:
    NodeAnchor, = jac_import('jac_cloud.core.architype', 'py', items={'NodeAnchor': None})

class get_channels_by_date(graph_walker, Walker):
    start_date: str = field('')
    end_date: str = field('')
    timezone: str = field('UTC')
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        start = datetime.strptime(f'{self.start_date}T00:00:00+00:00', '%Y-%m-%dT%H:%M:%S%z')
        end = datetime.strptime(f'{self.end_date}T00:00:00+00:00', '%Y-%m-%dT%H:%M:%S%z')
        end = end + timedelta(days=1) - timedelta(milliseconds=1)
        days = 1 if (end - start).days == 0 else (end - start).days
        try:
            collection = NodeAnchor.Collection.get_collection('interactions')
            pipeline = JacList([{'$match': {'agent_id': self.agent_id, 'time_stamp': {'$gte': self.start_date, '$lte': self.end_date}}}, {'$group': {'_id': {'$dateToString': {'format': '%Y-%m-%dT00:00:00.000Z' if days > 1 else '%Y-%m-%dT%H:00:00.000Z', 'date': {'$dateFromString': {'dateString': '$time_stamp'}}, 'timezone': self.timezone}}, 'unique_users': {'$addToSet': '$channel'}}}, {'$project': {'_id': 0, 'date': '$_id', 'count': {'$size': '$unique_users'}}}, {'$sort': {'date': 1}}])
            result = list(collection.aggregate(pipeline))
            total = sum(JacList([doc['count'] for doc in result]))
            Jac.report({'total': total, 'data': result})
        except Exception as e:
            self.logger.error(f'an exception occurred, {traceback.format_exc()}')


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from typing import Any, Optional
else:
    Any, Optional = jac_import('typing', 'py', items={'Any': None, 'Optional': None})
if typing.TYPE_CHECKING:
    from uuid import uuid4
else:
    uuid4, = jac_import('uuid', 'py', items={'uuid4': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from interaction import Interaction
else:
    Interaction, = jac_import('interaction', items={'Interaction': None})
if typing.TYPE_CHECKING:
    from interaction_response import InteractionResponse, TextInteractionMessage
else:
    InteractionResponse, TextInteractionMessage = jac_import('interaction_response', items={'InteractionResponse': None, 'TextInteractionMessage': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_node import GraphNode
else:
    GraphNode, = jac_import('jivas.agent.core.graph_node', items={'GraphNode': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_action import InteractAction
else:
    InteractAction, = jac_import('jivas.agent.action.interact_action', items={'InteractAction': None})
if typing.TYPE_CHECKING:
    from tail import Tail
else:
    Tail, = jac_import('tail', items={'Tail': None})
if typing.TYPE_CHECKING:
    from advance import Advance
else:
    Advance, = jac_import('advance', items={'Advance': None})
if typing.TYPE_CHECKING:
    from retrace import Retrace
else:
    Retrace, = jac_import('retrace', items={'Retrace': None})

class Frame(GraphNode, Node):
    agent_id: str = field('')
    session_id: str = field(gen=lambda: str(uuid4()))
    user_name: str = field('')
    variables: dict = field(gen=lambda: {})

    def variable_get(self, key: str, default: Optional[Any]=None) -> Any:
        if (variable := self.variables.get(key)):
            return variable
        elif default:
            return default
        else:
            return None

    def variable_set(self, key: str, value: Any) -> None:
        self.variables[key] = value

    def variable_del(self, key: str) -> None:
        if key in self.variables.keys():
            del self.variables[key]

    def get_last_interaction(self) -> Interaction:
        if (result := self.refs(Tail).filter(Interaction, None)):
            return Utils.node_obj(result)
        return None

    def is_first_interaction(self) -> bool:
        last_interaction = self.get_last_interaction()
        return 'new_user' in last_interaction.context_data

    def set_resume_action(self, action_label: str) -> None:
        if (last_interaction := self.get_last_interaction()):
            last_interaction.context_data['resume_action'] = action_label

    def get_resume_action(self, last_interaction_node: Interaction) -> InteractAction:
        action_node = None
        if last_interaction_node:
            if 'resume_action' in last_interaction_node.context_data:
                agent_node = self.get_agent()
                action_label = last_interaction_node.context_data['resume_action']
                action_node = agent_node.get_actions().get(action_label=action_label)
        return action_node

    def add_interaction(self, utterance: str, channel: str='default') -> Interaction:
        last_interaction_node = self.get_last_interaction()
        interaction_node = Interaction(agent_id=self.agent_id, channel=channel, utterance=utterance, response=InteractionResponse(session_id=self.session_id).export())
        return self.insert_interaction(interaction_node, last_interaction_node)

    def add_unprompted_interaction(self, message: str, channel: str='default') -> Interaction:
        last_interaction_node = self.get_last_interaction()
        interaction_node = Interaction(agent_id=self.agent_id, channel=channel, response=InteractionResponse(session_id=self.session_id, message=TextInteractionMessage(content=message).export()).export())
        return self.insert_interaction(interaction_node, last_interaction_node)

    def insert_interaction(self, interaction_node: Interaction, last_interaction_node: Interaction=None) -> Interaction:
        if not last_interaction_node and (not self.refs(Advance)):
            self.connect(interaction_node, edge=Advance)
            interaction_node.connect(self, edge=Retrace)
            self.connect(interaction_node, edge=Tail)
            interaction_node.context_data['new_user'] = True
        elif self.refs(Tail):
            self.disconnect(self.refs(Tail), edge=Tail)
            self.connect(interaction_node, edge=Tail)
            last_interaction_node.attach_interaction(interaction_node)
        return interaction_node

    def get_transcript_statements(self, interactions: int=10, max_statement_length: int=0, with_events: bool=False) -> list[dict]:
        return self.spawn(_get_transcript_statements(interactions, max_statement_length, with_events=with_events)).statements

    def get_transcript(self, interactions: int=10, max_statement_length: int=0, with_events: bool=False) -> str:
        transcript = ''
        statements = self.get_transcript_statements(interactions, max_statement_length, with_events=with_events)
        for item in statements:
            key = next(iter(item))
            value = item[key]
            transcript += f'{key} : {value} \\n'
        return transcript

    def get_interactions(self) -> list:
        return self.spawn(_get_interactions()).interactions

    def get_agent(self) -> None:
        return jobj(id=self.agent_id)

    def prune_interactions(self, frame_size: int) -> None:
        self.spawn(_prune_interactions(frame_size=frame_size))

    def refresh_interactions(self) -> None:
        self.prune_interactions(frame_size=1)

    def get_user_name(self, full: bool=False) -> None:
        if self.user_name and full:
            return self.user_name
        elif self.user_name:
            return Utils.extract_first_name(self.user_name)
        else:
            return 'user'

    def set_user_name(self, user_name: str) -> None:
        if user_name:
            self.user_name = user_name

class _get_transcript_statements(Walker):
    interactions: int = field(1)
    max_statement_length: int = field(0)
    statements: list = field(gen=lambda: JacList([]))
    last_interaction: Interaction = field(None)
    retrace_count: int = field(0)
    with_events: bool = field(False)

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_frame(self, here: Frame) -> None:
        last_interaction = here.get_last_interaction()
        self.visit(last_interaction)

    @with_entry
    def on_interaction(self, here: Interaction) -> None:
        if here == self.last_interaction:
            if not self.visit(here.refs(Retrace).filter(Interaction, None)):
                return self.disengage()
        elif self.retrace_count < self.interactions:
            ai_statement = ''
            human_statement = ''
            if here.events and self.with_events:
                for event in here.events:
                    self.statements.append({'ai': f'event: {event}'})
            if here.has_response():
                content = here.get_message().get_content()
                if type(content) == list:
                    for item in content:
                        ai_statement = item['content']
                        if ai_statement:
                            ai_statement = self.chunk_message(message=ai_statement)
                            self.statements.append({'ai': ai_statement})
                else:
                    ai_statement = self.chunk_message(message=content)
                    self.statements.append({'ai': ai_statement})
            if here.utterance:
                human_statement = Utils.escape_string(here.utterance)
                human_statement = self.chunk_message(message=human_statement)
                self.statements.append({'human': human_statement})
            self.retrace_count += 1
            self.visit(here.refs(Retrace).filter(Interaction, None))

    def chunk_message(self, message: str) -> str:
        if self.max_statement_length > 0:
            message_chunks = Utils.chunk_long_message(message=message, max_length=self.max_statement_length, chunk_length=self.max_statement_length)
            if len(message_chunks) > 1:
                message = message_chunks[0]
        return message

    @with_exit
    def on_exit(self, here) -> None:
        self.statements.reverse()

class _get_interactions(Walker):
    interactions: list = field(gen=lambda: JacList([]))

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_frame(self, here: Frame) -> None:
        self.visit(here.refs(Advance).filter(Interaction, None))

    @with_entry
    def on_interaction(self, here: Interaction) -> None:
        self.interactions.append(here)
        self.visit(here.refs(Advance).filter(Interaction, None))

class _prune_interactions(Walker):
    frame_size: int = field(1)
    frame_node: Frame = field(None)
    interaction_head: Interaction = field(None)
    retraces: int = field(0)

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_frame(self, here: Frame) -> None:
        self.frame_node = here
        if self.interaction_head:
            here.connect(self.interaction_head, edge=Advance)
        else:
            self.visit(here.get_last_interaction())

    @with_entry
    def on_interaction(self, here: Interaction) -> None:
        self.retraces += 1
        if self.retraces == self.frame_size and here.refs(Retrace).filter(Interaction, None):
            self.interaction_head = here
            self.visit(here.refs(Retrace).filter(Interaction, None))
            here.disconnect(here.refs(Retrace).filter(Interaction, None))
            if not here.refs(Retrace).filter(Frame, None):
                here.connect(self.frame_node, edge=Retrace)
        elif self.retraces > self.frame_size:
            if not self.visit(here.refs(Retrace).filter(Interaction, None)):
                del here
        else:
            self.visit(here.refs(Retrace).filter(Interaction, None))


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from graph_walker import graph_walker
else:
    graph_walker, = jac_import('graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from export_agent import export_agent
else:
    export_agent, = jac_import('export_agent', items={'export_agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})

class update_agent(graph_walker, Walker):
    agent_id: str = field('')
    agent_data: dict = field(gen=lambda: {})
    with_actions: bool = field(False)
    reporting: bool = field(True)
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        if (agent_node := here.update(data=self.agent_data, with_actions=self.with_actions)):
            if agent_node:
                if self.reporting:
                    Jac.report(agent_node.export_descriptor())
            else:
                self.logger.error('unable to update agent')
                if self.reporting:
                    Jac.get_context().status = 500
                    Jac.report('unable to update agent')


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from agents import Agents
else:
    Agents, = jac_import('agents', items={'Agents': None})
if typing.TYPE_CHECKING:
    from graph_walker import graph_walker
else:
    graph_walker, = jac_import('graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from import_agent import import_agent
else:
    import_agent, = jac_import('import_agent', items={'import_agent': None})

class init_agents(graph_walker, Walker):
    reporting: bool = field(False)
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agents(self, here: Agents) -> None:
        if (agent_nodes := here.get_all()):
            for agent_node in agent_nodes:
                try:
                    self.logger.info(f'initializing agent {agent_node.name}')
                    if not os.path.exists(agent_node.descriptor):
                        self.logger.error(f'agent descriptor not found: {agent_node.descriptor}')
                        continue
                    descriptor = ''
                    with open(agent_node.descriptor) as file:
                        descriptor = file.read()
                    if descriptor:
                        here.spawn(import_agent(descriptor=descriptor, reporting=self.reporting))
                except Exception as e:
                    self.logger.error(f'an exception occurred, {traceback.format_exc()}')


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from action import Action
else:
    Action, = jac_import('action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from interact_graph_walker import interact_graph_walker
else:
    interact_graph_walker, = jac_import('interact_graph_walker', items={'interact_graph_walker': None})

class get_action(interact_graph_walker, Walker):
    agent_id: str = field('')
    action_id: str = field('')
    action_data: dict = field(gen=lambda: {})
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Actions, None))

    @with_entry
    def on_actions(self, here: Actions) -> None:
        for action_node in here.get_all():
            if self.action_id == action_node.id:
                self.visit(action_node)

    @with_entry
    def on_action(self, here: Action) -> None:
        self.action_data = here.export()
        Jac.report(self.action_data)


"""Setup script for jivas"""

import os

from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version() -> str:
    """Get the package version from the __init__ file."""
    version_file = os.path.join(os.path.dirname(__file__), "jivas", "__init__.py")
    with open(version_file) as f:
        for line in f:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Version not found.")


setup(
    name="jivas",
    version=get_version(),
    description="JIVAS is an Agentic Framework for rapidly prototyping and deploying graph-based, AI solutions.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="TrueSelph Inc.",
    author_email="admin@trueselph.com",
    url="https://github.com/TrueSelph/jivas",
    keywords=["jivas"],
    packages=find_packages(
        include=[
            "jivas",
            "jivas.*",
        ]
    ),
    include_package_data=True,
    package_data={"jivas": []},
    python_requires=">=3.12.0",
    install_requires=[
        "jvcli>=2.0.3",
        "jvserve>=2.0.2",
        "pytz>=2024.2",
        "types-pytz>=2024.2.0.20241003",
        "schedule>=1.2.2",
        "openai>=1.51.0",
        "langchain>=0.3.3",
        "langchain-community>=0.3.2",
        "langchain-core>=0.3.10",
        "langchain-experimental>=0.3.2",
        "langchain-openai>=0.2.2",
        "setuptools>=75.8.1",
        "transformers>=4.49.0",
        "spacy>=3.8.4",
    ],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "pytest-mock",
            "pytest-cov",
            "coverage",
        ]
    },
)



"""Jivas Agent Utils Module"""

import ast
import json
import logging
import mimetypes
import os
import re
import time
from collections import defaultdict, deque
from datetime import datetime
from typing import Any, DefaultDict, Dict, List, Optional
from uuid import UUID

import pytz  # To handle timezones
import requests
import yaml

logger = logging.getLogger(__name__)


class LongStringDumper(yaml.SafeDumper):
    """Custom YAML dumper to handle long strings."""

    def represent_scalar(
        self, tag: str, value: str, style: str | None = None
    ) -> yaml.nodes.ScalarNode:
        """Represent scalar values in YAML."""
        # Replace any escape sequences to format the output as desired
        if (
            len(value) > 150 or "\n" in value
        ):  # Adjust the threshold for long strings as needed
            style = "|"
            # converts all newline escapes to actual representations
            value = "\n".join([line.rstrip() for line in value.split("\n")])
        else:
            # converts all newline escapes to actual representations
            value = "\n".join([line.rstrip() for line in value.split("\n")]).rstrip()

        return super().represent_scalar(tag, value, style)


class Utils:
    """Utility class for various helper methods."""

    @staticmethod
    def get_descriptor_root() -> str:
        """Get the root path for descriptors."""

        descriptor_path = os.environ.get("JIVAS_DESCRIPTOR_ROOT_PATH", ".jvdata")

        if not os.path.exists(descriptor_path):
            os.makedirs(descriptor_path)

        return descriptor_path

    @staticmethod
    def get_actions_root() -> str:
        """Get the root path for actions."""

        actions_root_path = os.environ.get("JIVAS_ACTIONS_ROOT_PATH", "actions")

        if not os.path.exists(actions_root_path):
            os.makedirs(actions_root_path)

        return actions_root_path

    @staticmethod
    def get_daf_root() -> str:
        """Get the root path for DAF."""

        daf_root_path = os.environ.get("JIVAS_DAF_ROOT_PATH", "daf")

        if not os.path.exists(daf_root_path):
            os.makedirs(daf_root_path)

        return daf_root_path

    @staticmethod
    def dump_yaml_file(file_path: str, data: dict) -> None:
        """Dump data to a YAML file."""
        try:
            with open(file_path, "w") as file:
                yaml_output = yaml.dump(
                    data,
                    Dumper=LongStringDumper,
                    allow_unicode=True,
                    default_flow_style=False,
                    sort_keys=False,
                )
                file.write(yaml_output)
            logger.debug(f"Descriptor successfully written to {file_path}")
        except IOError:
            logger.error(f"Error writing to descriptor file {file_path}")

    @staticmethod
    def path_to_module(path: str) -> str:
        """
        Converts a file path into a module path.

        Parameters:
            path (str): The file path to convert. Example: '/a/b/c'

        Returns:
            str: The module path. Example: 'a.b.c'
        """
        # Strip leading and trailing slashes and split the path by slashes
        parts = path.strip("/").split("/")

        # Join the parts with dots to form the module path
        module_path = ".".join(parts)

        return module_path

    @staticmethod
    def find_package_folder(
        rootdir: str, name: str, required_files: set | None = None
    ) -> str | None:
        """Find a package folder within a namespace."""
        if required_files is None:
            required_files = {"info.yaml"}
        try:
            # Split the provided name into namespace and package_folder
            namespace, package_folder = name.split("/")

            # Build the path for the namespace
            namespace_path = os.path.join(rootdir, namespace)

            # Check if the namespace directory exists
            if not os.path.isdir(namespace_path):
                return None

            # Traverse the namespace directory for the package_folder
            for root, dirs, _files in os.walk(namespace_path):
                if package_folder in dirs:
                    package_path = os.path.join(root, package_folder)

                    # Check for the presence of the required files
                    folder_files = set(os.listdir(package_path))

                    if required_files.issubset(folder_files):
                        # Return the full path of the package_folder if all checks out
                        return package_path

            # If package_folder is not found, return None
            return None

        except ValueError:
            logger.error("Invalid format. Please use 'namespace/package_folder'.")
            return None

    @staticmethod
    def list_to_phrase(lst: list) -> str:
        """Formats a list as a phrased list, e.g. ['one', 'two', 'three'] becomes 'one, two, and three'."""
        if not lst:
            return ""

        # Convert all elements to strings
        lst = list(map(str, lst))

        if len(lst) == 1:
            return lst[0]

        return ", ".join(lst[:-1]) + f", and {lst[-1]}"

    @staticmethod
    def replace_placeholders(
        string_or_collection: str | list, placeholders: dict
    ) -> str | list | None:
        """
        Replaces placeholders delimited by {{ and }} in a string or collection of strings with their corresponding values
        from a dictionary of key-value pairs. If a value in the dictionary is a list, it is formatted as a phrased list.
        Returns only items that have no remaining placeholders.

        Parameters:
        - string_or_collection (str or list): A string or collection of strings containing placeholders delimited by {{ and }}.
        - placeholders (dict): A dictionary of key-value pairs where the keys correspond to the placeholder names inside the {{ and }}
                                and the values will replace the entire placeholder in the string or collection of strings. If a
                                value in the dictionary is a list, it will be formatted as a phrased list.

        Returns:
        - str or list: The input string or collection of strings with the placeholders replaced by their corresponding values.
                    Returns only items that have no remaining placeholders.
        """

        def replace_in_string(s: str, placeholders: dict) -> str | None:
            # Replace placeholders in a single string
            for key, value in placeholders.items():
                if isinstance(value, list):
                    value = Utils.list_to_phrase(value)
                s = s.replace(f"{{{{{key}}}}}", str(value))
            return s if not re.search(r"{{.*?}}", s) else None

        if isinstance(string_or_collection, str):
            return replace_in_string(string_or_collection, placeholders)

        elif isinstance(string_or_collection, list):
            # Process each string in the list and return those with no placeholders left
            return [
                result
                for string in string_or_collection
                if (result := replace_in_string(string, placeholders)) is not None
            ]

        else:
            raise TypeError("Input must be a string or a list of strings.")

    @staticmethod
    def chunk_long_message(
        message: str, max_length: int = 1024, chunk_length: int = 1024
    ) -> list:
        """
        Splits a long message into smaller chunks of no more than chunk_length characters,
        ensuring no single chunk exceeds max_length.
        """
        if len(message) <= max_length:
            return [message]

        # Initialize variables
        final_chunks = []
        current_chunk = ""
        current_chunk_length = 0

        # Split the message into words while preserving newline characters
        words = re.findall(r"\S+\n*|\n+", message)
        words = [word for word in words if word.strip()]  # Filter out empty strings

        for word in words:
            word_length = len(word)

            if current_chunk_length + word_length + 1 <= chunk_length:
                # Add the word to the current chunk
                if current_chunk:
                    current_chunk += " "
                current_chunk += word
                current_chunk_length += word_length + 1
            else:
                # If the current chunk is full, add it to the list of chunks
                final_chunks.append(current_chunk)
                current_chunk = word  # Start a new chunk with the current word
                current_chunk_length = word_length

        if current_chunk:
            # Add the last chunk if it's non-empty
            final_chunks.append(current_chunk)

        return final_chunks

    @staticmethod
    def escape_string(input_string: str) -> str:
        """
        Escapes curly braces in the input string by converting them to double curly braces.

        Args:
        - input_string (str): The string to be escaped.

        Returns:
        - str: The escaped string with all curly braces doubled.
        """
        if not isinstance(input_string, (str)):
            logger.error(f"Error expect string: {input_string}")
            return ""
        else:
            return input_string.replace("{", "{{").replace("}", "}}")

    @staticmethod
    def export_to_dict(data: object | dict, ignore_keys: list | None = None) -> dict:
        """Export an object to a dictionary, ignoring specified keys."""
        if ignore_keys is None:
            ignore_keys = ["__jac__"]

        def stringify_value(value: object) -> object:
            # Recursive handling of dictionaries and lists
            if isinstance(value, dict):
                return {
                    k: stringify_value(v)
                    for k, v in value.items()
                    if k not in ignore_keys
                }
            elif isinstance(value, list):
                return [stringify_value(item) for item in value]
            elif isinstance(value, (str, int, float, bool, type(None))):
                return value
            else:
                # Stringify any other complex objects
                return str(value)

        # Convert top-level object to dictionary if possible
        if hasattr(data, "__dict__"):
            data = data.__dict__

        if isinstance(data, dict):
            return {
                k: stringify_value(v) for k, v in data.items() if k not in ignore_keys
            }
        else:
            return {}

    @staticmethod
    def safe_json_dump(data: dict) -> str | None:
        """Safely convert a dictionary with mixed types to a JSON string for logs."""

        def serialize(obj: dict) -> dict:
            """Recursively convert strings within complex objects."""

            def wrap_content(value: object) -> object:
                # Return value wrapped in a dictionary with key 'content' if it's a str, int or float
                return (
                    {"content": value}
                    if isinstance(value, (str, int, float))
                    else value
                )

            def process_dict(d: dict) -> dict:
                for key, value in d.items():
                    if isinstance(value, dict):
                        d[key] = process_dict(value)
                    elif isinstance(value, list):
                        # If the list contains dictionaries, recursively process each
                        d[key] = [
                            process_dict(item) if isinstance(item, dict) else item
                            for item in value
                        ]
                    else:
                        d[key] = wrap_content(value)
                return d

            # Create a deep copy of the original dict to avoid mutation
            import copy

            result = copy.deepcopy(obj)

            return process_dict(result)

        try:
            # Attempt to serialize the dictionary
            return json.dumps(serialize(data))
        except (TypeError, ValueError) as e:
            # Handle serialization errors
            logger.error(f"Serialization error: {str(e)}")
            return None

    @staticmethod
    def date_now(
        timezone: str = "US/Eastern", date_format: str = "%Y-%m-%d"
    ) -> str | None:
        """Get the current date and time in the specified timezone and format."""
        try:
            # If a timezone is specified, apply it
            tz = pytz.timezone(timezone)
            # Get the current datetime
            dt = datetime.now(tz)
            # Format the datetime according to the provided format
            formatted_datetime = dt.strftime(date_format)

            return formatted_datetime
        except Exception:
            return None

    @staticmethod
    def is_valid_uuid(uuid_to_test: str, version: int = 4) -> bool:
        """
        Check if uuid_to_test is a valid UUID.

        Parameters
        ----------
        uuid_to_test : str
        version : {1, 2, 3, 4}

        Returns
        -------
        `True` if uuid_to_test is a valid UUID, otherwise `False`.

        Examples
        --------
        >>> is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a')
        True
        >>> is_valid_uuid('c9bf9e58')
        False
        """

        try:
            uuid_obj = UUID(uuid_to_test, version=version)
        except ValueError:
            return False
        return str(uuid_obj) == uuid_to_test

    @staticmethod
    def node_obj(node_list: list) -> object | None:
        """Return the first object in the list or None if the list is empty."""
        return node_list[0] if node_list else None

    @staticmethod
    def order_interact_actions(
        actions_data: List[Dict[str, Any]]
    ) -> Optional[List[Dict[str, Any]]]:
        """Order interact actions based on their dependencies and weights."""
        if not actions_data:
            return None

        other_actions = [
            action
            for action in actions_data
            if action.get("context", {}).get("_package", {}).get("meta", {}).get("type")
            != "interact_action"
        ]
        interact_actions = [
            action
            for action in actions_data
            if action.get("context", {}).get("_package", {}).get("meta", {}).get("type")
            == "interact_action"
        ]

        action_lookup = {
            action["context"]["_package"]["name"]: action for action in interact_actions
        }

        graph: DefaultDict[str, List[str]] = defaultdict(list)
        in_degree: DefaultDict[str, int] = defaultdict(int)

        action_weights = {
            name: action["context"]["_package"]["config"]
            .get("order", {})
            .get("weight", 0)
            for name, action in action_lookup.items()
        }

        # First handle explicit before/after constraints (excluding "all")
        for action in interact_actions:
            action_name = action["context"]["_package"]["name"]
            config_order = action["context"]["_package"]["config"].get("order", {})

            before = config_order.get("before")
            after = config_order.get("after")

            if before and before != "all" and before in action_lookup:
                graph[action_name].append(before)
                in_degree[before] += 1

            if after and after != "all" and after in action_lookup:
                graph[after].append(action_name)
                in_degree[action_name] += 1

        # Handle "before": "all" and "after": "all" constraints separately without creating cycles
        before_all_actions = [
            action["context"]["_package"]["name"]
            for action in interact_actions
            if action["context"]["_package"]["config"].get("order", {}).get("before")
            == "all"
        ]
        after_all_actions = [
            action["context"]["_package"]["name"]
            for action in interact_actions
            if action["context"]["_package"]["config"].get("order", {}).get("after")
            == "all"
        ]

        for action_name in before_all_actions:
            for other_name in action_lookup:
                if other_name not in before_all_actions and other_name != action_name:
                    graph[action_name].append(other_name)
                    in_degree[other_name] += 1

        for action_name in after_all_actions:
            for other_name in action_lookup:
                if other_name not in after_all_actions and other_name != action_name:
                    graph[other_name].append(action_name)
                    in_degree[action_name] += 1

        # Kahn's algorithm with weights as tie-breaker
        queue = deque(
            sorted(
                [name for name in action_lookup if in_degree[name] == 0],
                key=lambda x: action_weights[x],
            )
        )

        sorted_actions_names = []
        while queue:
            current = queue.popleft()
            sorted_actions_names.append(current)
            for neighbor in graph[current]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    queue.append(neighbor)
            queue = deque(sorted(queue, key=lambda x: action_weights[x]))

        if len(sorted_actions_names) != len(interact_actions):
            raise ValueError("Circular dependency detected!")

        # Map sorted names back to actions and add others
        sorted_actions = [
            action_lookup[name] for name in sorted_actions_names
        ] + other_actions

        # Assign final weights according to sorted order
        for idx, action in enumerate(sorted_actions):
            if action["context"]["_package"]["meta"]["type"] == "interact_action":
                action["context"]["weight"] = idx

        return sorted_actions

    @staticmethod
    def get_mime_type(
        file_path: str | None = None,
        url: str | None = None,
        mime_type: str | None = None,
    ) -> dict | None:
        """Determines the MIME type of a file or URL and categorizes it into common file types (image, document, audio, video)."""
        detected_mime_type = None

        if file_path:
            # Use mimetypes to guess MIME type based on file extension
            detected_mime_type, _ = mimetypes.guess_type(file_path)
        elif url:
            # Make a HEAD request to get the Content-Type header
            try:
                response = requests.head(url, allow_redirects=True)
                detected_mime_type = response.headers.get("Content-Type")
            except requests.RequestException as e:
                logger.error(f"Error making HEAD request: {e}")
                return None
        else:
            # Fallback to initial MIME type if provided
            detected_mime_type = mime_type

        # MIME type categories
        image_mime_types = [
            "image/jpeg",
            "image/png",
            "image/gif",
            "image/bmp",
            "image/webp",
            "image/tiff",
            "image/svg+xml",
            "image/x-icon",
            "image/heic",
            "image/heif",
            "image/x-raw",
        ]
        document_mime_types = [
            "application/pdf",
            "application/msword",
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            "application/vnd.ms-excel",
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "application/vnd.ms-powerpoint",
            "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            "text/plain",
            "text/csv",
            "text/html",
            "application/rtf",
            "application/x-tex",
            "application/vnd.oasis.opendocument.text",
            "application/vnd.oasis.opendocument.spreadsheet",
            "application/epub+zip",
            "application/x-mobipocket-ebook",
            "application/x-fictionbook+xml",
            "application/x-abiword",
            "application/vnd.apple.pages",
            "application/vnd.google-apps.document",
        ]
        audio_mime_types = [
            "audio/mpeg",
            "audio/wav",
            "audio/ogg",
            "audio/flac",
            "audio/aac",
            "audio/mp3",
            "audio/webm",
            "audio/amr",
            "audio/midi",
            "audio/x-m4a",
            "audio/x-realaudio",
            "audio/x-aiff",
            "audio/x-wav",
            "audio/x-matroska",
        ]
        video_mime_types = [
            "video/mp4",
            "video/mpeg",
            "video/ogg",
            "video/webm",
            "video/quicktime",
            "video/x-msvideo",
            "video/x-matroska",
            "video/x-flv",
            "video/x-ms-wmv",
            "video/3gpp",
            "video/3gpp2",
            "video/h264",
            "video/h265",
            "video/x-f4v",
            "video/avi",
        ]

        # Handle cases where MIME type cannot be detected
        if detected_mime_type is None or detected_mime_type == "binary/octet-stream":
            if file_path:
                _, file_extension = os.path.splitext(file_path)
            elif url:
                _, file_extension = os.path.splitext(url)
            else:
                file_extension = ""

            detected_mime_type = mimetypes.types_map.get(file_extension.lower())

        # Categorize MIME type
        if detected_mime_type in image_mime_types:
            return {"file_type": "image", "mime": detected_mime_type}
        elif detected_mime_type in document_mime_types:
            return {"file_type": "document", "mime": detected_mime_type}
        elif detected_mime_type in audio_mime_types:
            return {"file_type": "audio", "mime": detected_mime_type}
        elif detected_mime_type in video_mime_types:
            return {"file_type": "video", "mime": detected_mime_type}
        else:
            logger.error(f"Unsupported MIME Type: {detected_mime_type}")
            return None

    @staticmethod
    def convert_str_to_json(text: str) -> dict | None:
        """Convert a string to a JSON object."""
        if isinstance(text, str):
            text = text.replace("```json", "")
            text = text.replace("```", "")
        try:
            if isinstance(text, (dict, list)):
                return text
            else:
                return json.loads(text)
        except (json.JSONDecodeError, TypeError):
            try:
                return ast.literal_eval(text)
            except (SyntaxError, ValueError) as e:
                if "'{' was never closed" in str(e):
                    text = text + "}"
                    return json.loads(text)
                else:
                    logger.error(e)
                    return None

    @staticmethod
    def delete_files(
        directory: str, days: int = 30, filenames_to_delete: list | None = None
    ) -> None:
        """Delete files in a directory older than a specified number of days or matching specified filenames."""
        if filenames_to_delete is None:
            filenames_to_delete = []
        # Get the current time
        current_time = time.time()
        try:
            # List all files in the directory
            for filename in os.listdir(directory):
                file_path = os.path.join(directory, filename)

                # Check if it's a file
                if os.path.isfile(file_path):
                    # Get the file's last modified time
                    file_mod_time = os.path.getmtime(file_path)

                    # Check if the file is older than the specified number of days or matches a filename to delete
                    if (current_time - file_mod_time > days * 86400) or (
                        filename in filenames_to_delete
                    ):
                        try:
                            # Delete the file
                            os.remove(file_path)
                            print(f"Deleted: {file_path}")
                        except Exception as e:
                            print(f"Failed to delete {file_path}: {e}")

        except Exception as e:
            print(f"Error deleting files: {e}")

    @staticmethod
    def extract_first_name(full_name: str) -> str:
        """Extract the first name from a full name."""
        # List of common titles to be removed
        titles = ["Mr.", "Mrs.", "Ms.", "Miss", "Dr.", "Prof.", "Sir", "Madam", "Mx."]

        # Split the full name into parts
        name_parts = full_name.split()

        # Remove any titles from the list of name parts
        name_parts = [part for part in name_parts if part not in titles]

        # Return the first element as the first name if it's available
        if name_parts:
            return name_parts[0]
        else:
            return ""
